@@ -26,6 +26,7 @@ import os
 import subprocess
 import sys
 import requests
+import io
 if sys.version_info[0] == 2:
     import httplib
 elif sys.version_info[0] == 3:
@@ -38,15 +39,17 @@ logger = logging.getLogger(__name__)
 
 # Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
 # The only method currently relying on it is s3_iter_bucket, which is instructed
-# not to use it by the NO_MULTIPROCESSING flag.
+# whether to use it by the MULTIPROCESSING flag.
+MULTIPROCESSING = False
 try:
     import multiprocessing.pool
+    MULTIPROCESSING = True
 except ImportError:
     logger.warning("multiprocessing could not be imported and won't be used")
-    NO_MULTIPROCESSING = True
     from itertools import imap
-else:
-    NO_MULTIPROCESSING = False
+
+from . import gzipstreamfile
+
 
 S3_MIN_PART_SIZE = 50 * 1024**2  # minimum part size for S3 multipart uploads
 WEBHDFS_MIN_PART_SIZE = 50 * 1024**2  # minimum part size for HDFS multipart uploads
@@ -101,6 +104,11 @@ def smart_open(uri, mode="rb", **kw):
       ...    fout.write("hello world!\n")
       >>> with smart_open.smart_open('/home/radim/another.txt.bz2', 'wb') as fout:
       ...    fout.write("good bye!\n")
+      >>> # stream from/to (compressed) local files with Expand ~ and ~user constructions:
+      >>> for line in smart_open.smart_open('~/my_file.txt'):
+      ...    print line
+      >>> for line in smart_open.smart_open('my_file.txt'):
+      ...    print line
 
     """
 
@@ -186,6 +194,8 @@ class ParseUri(object):
       * hdfs://path/file
       * webhdfs://host:port/path/file
       * ./local/path/file
+      * ~/local/path/file
+      * local/path/file
       * ./local/path/file.gz
       * file:///home/user/file
       * file:///home/user/file.bz2
@@ -212,6 +222,8 @@ class ParseUri(object):
                 raise RuntimeError("invalid HDFS URI: %s" % uri)
         elif self.scheme == "webhdfs":
             self.uri_path = parsed_uri.netloc + "/webhdfs/v1" + parsed_uri.path
+            if parsed_uri.query:
+                self.uri_path += "?" + parsed_uri.query
 
             if not self.uri_path:
                 raise RuntimeError("invalid WebHDFS URI: %s" % uri)
@@ -242,45 +254,163 @@ class ParseUri(object):
         elif self.scheme == 'file':
             self.uri_path = parsed_uri.netloc + parsed_uri.path
 
+            # '~/tmp' may be expanded to '/Users/username/tmp'
+            self.uri_path = os.path.expanduser(self.uri_path)
+
             if not self.uri_path:
                 raise RuntimeError("invalid file URI: %s" % uri)
         else:
             raise NotImplementedError("unknown URI scheme %r in %r" % (self.scheme, uri))
 
 
+def is_gzip(name):
+    """Return True if the name indicates that the file is compressed with
+    gzip."""
+    return name.endswith(".gz")
+
+
+class S3ReadStreamInner(object):
+
+    def __init__(self, stream):
+        self.stream = stream
+        self.unused_buffer = b''
+        self.closed = False
+        self.finished = False
+
+    def read_until_eof(self):
+        #
+        # This method is here because boto.s3.Key.read() reads the entire
+        # file, which isn't expected behavior.
+        #
+        # https://github.com/boto/boto/issues/3311
+        #
+        buf = b""
+        while not self.finished:
+            raw = self.stream.read(io.DEFAULT_BUFFER_SIZE)
+            if len(raw) > 0:
+                buf += raw
+            else:
+                self.finished = True
+        return buf
+
+    def read_from_buffer(self, size):
+        """Remove at most size bytes from our buffer and return them."""
+        part = self.unused_buffer[:size]
+        self.unused_buffer = self.unused_buffer[size:]
+        return part
+
+    def read(self, size=None):
+        if not size or size < 0:
+            return self.read_from_buffer(
+                len(self.unused_buffer)) + self.read_until_eof()
+
+        # Use unused data first
+        if len(self.unused_buffer) >= size:
+            return self.read_from_buffer(size)
+
+        # If the stream is finished and no unused raw data, return what we have
+        if self.stream.closed or self.finished:
+            self.finished = True
+            return self.read_from_buffer(size)
+
+        # Consume new data in chunks and return it.
+        while len(self.unused_buffer) < size:
+            raw = self.stream.read(io.DEFAULT_BUFFER_SIZE)
+            if len(raw):
+                self.unused_buffer += raw
+            else:
+                self.finished = True
+                break
+
+        return self.read_from_buffer(size)
+
+    def readinto(self, b):
+        # Read up to len(b) bytes into bytearray b
+        # Sadly not as efficient as lower level
+        data = self.read(len(b))
+        if not data:
+            return None
+        b[:len(data)] = data
+        return len(data)
+
+    def readable(self):
+        # io.BufferedReader needs us to appear readable
+        return True
+
+    def _checkReadable(self, msg=None):
+        # This is required to satisfy io.BufferedReader on Python 2.6.
+        # Another way to achieve this is to inherit from io.IOBase, but that
+        # leads to other problems.
+        return True
+
+
+class S3ReadStream(io.BufferedReader):
+
+    def __init__(self, key):
+        self.stream = S3ReadStreamInner(key)
+        super(S3ReadStream, self).__init__(self.stream)
+
+    def read(self, *args, **kwargs):
+        # Patch read to return '' instead of raise Value Error
+        # TODO: what actually raises ValueError in the following code?
+        try:
+            #
+            # io.BufferedReader behaves differently to a built-in file object.
+            # If the object is in non-blocking mode and no bytes are available,
+            # the former will return None. The latter returns an empty string.
+            # We want to behave like a built-in file object here.
+            #
+            result = super(S3ReadStream, self).read(*args, **kwargs)
+            if result is None:
+                return ""
+            return result
+        except ValueError:
+            return ''
+
+    def readline(self, *args, **kwargs):
+        # Patch readline to return '' instead of raise Value Error
+        # TODO: what actually raises ValueError in the following code?
+        try:
+            result = super(S3ReadStream, self).readline(*args, **kwargs)
+            return result
+        except ValueError:
+            return ''
+
+
 class S3OpenRead(object):
     """
     Implement streamed reader from S3, as an iterable & context manager.
 
+    Supports reading from gzip-compressed files.  Identifies such files by
+    their extension.
+
     """
     def __init__(self, read_key):
         if not hasattr(read_key, "bucket") and not hasattr(read_key, "name") and not hasattr(read_key, "read") \
                 and not hasattr(read_key, "close"):
             raise TypeError("can only process S3 keys")
         self.read_key = read_key
-        self.line_generator = s3_iter_lines(self.read_key)
+        self._open_reader()
+
+    def _open_reader(self):
+        if is_gzip(self.read_key.name):
+            self.reader = gzipstreamfile.GzipStreamFile(self.read_key)
+        else:
+            self.reader = S3ReadStream(self.read_key)
 
     def __iter__(self):
-        key = self.read_key.bucket.get_key(self.read_key.name)
-        if key is None:
-            raise KeyError(self.read_key.name)
+        for line in self.reader:
+            yield line
 
-        return s3_iter_lines(key)
+    def readline(self):
+        return self.reader.readline()
 
     def read(self, size=None):
         """
         Read a specified number of bytes from the key.
 
-        Note read() and line iteration (`for line in self: ...`) each have their
-        own file position, so they are independent. Doing a `read` will not affect
-        the line iteration, and vice versa.
-
         """
-        if not size or size < 0:
-            # For compatibility with standard Python, `read(negative)` = read the rest of the file.
-            # Otherwise, boto would read *from the start* if given size=-1.
-            size = 0
-        return self.read_key.read(size)
+        return self.reader.read(size)
 
     def seek(self, offset, whence=0):
         """
@@ -292,12 +422,13 @@ class S3OpenRead(object):
         if whence != 0 or offset != 0:
             raise NotImplementedError("seek other than offset=0 not implemented yet")
         self.read_key.close(fast=True)
+        self._open_reader()
 
     def __enter__(self):
         return self
 
     def __exit__(self, type, value, traceback):
-        self.read_key.close()
+        self.read_key.close(fast=True)
 
     def __str__(self):
         return "%s<key: %s>" % (
@@ -442,6 +573,9 @@ class S3OpenWrite(object):
         if not hasattr(outkey, "bucket") and not hasattr(outkey, "name"):
             raise TypeError("can only process S3 keys")
 
+        if is_gzip(outkey.name):
+            raise NotImplementedError("streaming write to S3 gzip not supported")
+
         self.outkey = outkey
         self.min_part_size = min_part_size
 
@@ -631,7 +765,7 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
 
     The keys are processed in parallel, using `workers` processes (default: 16),
     to speed up downloads greatly. If multiprocessing is not available, thus
-    NO_MULTIPROCESSING is True, this parameter will be ignored.
+    MULTIPROCESSING is False, this parameter will be ignored.
 
     Example::
 
@@ -649,13 +783,13 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
     total_size, key_no = 0, -1
     keys = (key for key in bucket.list(prefix=prefix) if accept_key(key.name))
 
-    if NO_MULTIPROCESSING:
-        logger.info("iterating over keys from %s without multiprocessing" % bucket)
-        iterator = imap(s3_iter_bucket_process_key, keys)
-    else:
+    if MULTIPROCESSING:
         logger.info("iterating over keys from %s with %i workers" % (bucket, workers))
         pool = multiprocessing.pool.Pool(processes=workers)
         iterator = pool.imap_unordered(s3_iter_bucket_process_key, keys)
+    else:
+        logger.info("iterating over keys from %s without multiprocessing" % bucket)
+        iterator = imap(s3_iter_bucket_process_key, keys)
 
     for key_no, (key, content) in enumerate(iterator):
         if key_no % 1000 == 0:
@@ -670,45 +804,12 @@ def s3_iter_bucket(bucket, prefix='', accept_key=lambda key: True, key_limit=Non
             # we were asked to output only a limited number of keys => we're done
             break
 
-    if not NO_MULTIPROCESSING:
+    if MULTIPROCESSING:
         pool.terminate()
 
     logger.info("processed %i keys, total size %i" % (key_no + 1, total_size))
 
 
-def s3_iter_lines(key):
-    """
-    Stream an object from S3 line by line (generator).
-
-    `key` must be a `boto.key.Key` object.
-
-    """
-    # check valid object on input
-    if not isinstance(key, boto.s3.key.Key):
-        raise TypeError("expected boto.key.Key object on input")
-
-    buf = b''
-    # keep reading chunks of bytes into the buffer
-    for chunk in key:
-        buf += chunk
-
-        start = 0
-        # process all lines within the current buffer
-        while True:
-            end = buf.find(b'\n', start) + 1
-            if end:
-                yield buf[start : end]
-                start = end
-            else:
-                # no more newlines => break out to read more data from s3 into the buffer
-                buf = buf[start : ]
-                break
-
-    # process the last line, too
-    if buf:
-        yield buf
-
-
 class WebHdfsException(Exception):
     def __init__(self, msg=str()):
         self.msg = msg
